/* FindRef.c
Originaly written by Anders Melchiorsen, modified by Roland Florac (SAS/C compiler) */

#include "FetchRefs.h"

static LONG __regargs BuildRefList (struct FindRefOptions *args);
static void __regargs FindRefInteractive (struct FindRefOptions *args, struct FindRefReturnStruct *ret);
static void __regargs PutRef (struct ListViewNode *item, STRPTR destfile, struct FindRefReturnStruct *ret);
static void SwitchToFileIndex (void);
static struct Screen * __regargs FindScreen (STRPTR pubname);

extern struct IntuitionBase * IntuitionBase;

APTR ListPool;
static struct List RefsFound;
static struct TR_Project *RefListPrj;

struct ListViewNode {

/* ln_name in the Node points to the string that is to be shown in the
 * listview. ln_type is 1 for file references, 2 for normal ones.
 *
 * fileentry is a pointer to the FileEntry in which the reference is
 *
 * refsentry is a pointer to the reference to fetch.
 *
 * ListViewText is used if we build a new string for the listview.
 */

    struct Node node;
    struct FileEntry *fileentry;
    struct RefsEntry *refsentry;
    UBYTE ListViewText[0];
};

/* Triton specifications */
enum {	ID_LIST_WINDOW = 1,
	ID_LIST_LIST, ID_LIST_PATTERN, ID_LIST_OK, ID_LIST_FILE, ID_LIST_CANCEL
     };

struct TagItem RefListOutline[] =
{
    WindowID (ID_LIST_WINDOW),
	WindowTitle (0),
	ScreenTitle (0),
	WindowBackfillWin,
	WindowPosition (TRWP_CENTERDISPLAY),
	WindowFlags (TRWF_HELP|TRWF_NOZIPGADGET|TRWF_NOESCCLOSE),
	TRWI_PubScreen, NULL,	     /* Filled in on the fly */

    HorizGroupA,
	Space,

	VertGroupA,
	    Space,

	    HorizGroupA,
		ListSS (&RefsFound, ID_LIST_LIST, 0, 0),
		    TRAT_MinWidth,  60,
		    TRAT_MinHeight, 15,
	    EndGroup,

	    Space,

	    HorizGroupC,
		TextN (0),
		    TRAT_ID, ID_LIST_PATTERN,
		Space,
		StringGadget ("", ID_LIST_PATTERN),
		    TRAT_Flags, TRST_NORETURNBROADCAST,
	    EndGroup,

	    Space,

	    HorizGroupE,
		ButtonR (0, ID_LIST_OK),
		Space,
		Button (0, ID_LIST_FILE),
		Space,
		ButtonE (0, ID_LIST_CANCEL),
	    EndGroup,

	    Space,
	EndGroup,

	Space,
    EndGroup,
EndProject
};

/* FindRef() - interface to fetch a reference to file
 * Actual look-up routine. Copies the reference to 'ref' into a file */
void __regargs FindRef (struct FindRefOptions *args, struct FindRefReturnStruct *ret)
{
    LONG refsfound = 0;
    UBYTE pattern[128];

    /* Assume failure until proven wrong */
    ret->Result = RET_FAULT;
    ret->Number = ERROR_NO_FREE_STORE;

    if (! NewMatchList())
	goto error;
    /* Build a list of references matching the requested one */

    /* If the search  pattern  turns out to be blank we show
     * the "references found" window anyway -- to allow the
     * user to enter a new one. */

    if (ParsePatternNoCase (args->Reference, pattern, 128) != -1)
	if (MatchPatternNoCase (pattern, ""))
	    refsfound = 2;

    if (!refsfound)
    {
	refsfound = BuildRefList (args);

	if (refsfound > 0)
	{
	    ret->Result = refsfound;
	    goto error;
	}
	else
	    refsfound = -refsfound;
    }

    /* If exactly one match was found we simply fetch it. Otherwise we
     * ask the user for further help.
     */
    if (refsfound == 1)
	PutRef ((struct ListViewNode *) RefsFound.lh_Head, args->DestFile, ret);
    else if (refsfound > 1)
	FindRefInteractive (args, ret);
    else
	ret->Result = RET_NO_MATCH;

error:

    /* Free the list */
    LibDeletePool (ListPool);
    ListPool = NULL;
}

/* FindRefInteractive() - open window and ask which reference to use */
static void __regargs FindRefInteractive (struct FindRefOptions *args, struct FindRefReturnStruct *ret)
{
    struct Screen *scr;
    BOOL swappedscreen = FALSE;

    /* Get a pointer to the Screen to open on. The screen is
     * LockPubScreen()'ed and we must unlock (is done further down).
     */
    if (!(scr = FindScreen (args->PubScreen)))
	return;

    /* Bring the screen to the front if it is not already there */
    if (IntuitionBase->FirstScreen != scr)
    {	ScreenToFront (scr);
	swappedscreen = TRUE;
    }

    /* Fill in the variable value of the TagList */
    {
	ULONG *t = (ULONG *) RefListOutline;

	while (*t++ != TRWI_PubScreen) ;
	*t = (ULONG)scr;
    }

    /* Open and handle everything */
    RefListOutline[1].ti_Data = (ULONG) GetString (TEXTE_REFERENCE);
    RefListOutline[2].ti_Data = (ULONG) GetString (TEXTE_VERSION);
    RefListOutline[23].ti_Data = (ULONG) GetString (TEXTE_FIND_PATTERN);
    RefListOutline[33].ti_Data = (ULONG) GetString (TEXTE_FETCHREF);
    RefListOutline[38].ti_Data = (ULONG) GetString (TEXTE_LISTE);
    RefListOutline[42].ti_Data = (ULONG) GetString (TEXTE_CANCEL);
    if (RefListPrj = TR_OpenProject (Application, RefListOutline))
    {
	struct TR_Message *msg;

	if (IsListEmpty (&RefsFound))
	{
	    /* Disable fetch gadgets if the references window is empty */
	    TR_SetAttribute (RefListPrj, ID_LIST_FILE, TRAT_Disabled, TRUE);
	    TR_SetAttribute (RefListPrj, ID_LIST_OK, TRAT_Disabled, TRUE);
	    TR_SendMessage (RefListPrj, ID_LIST_PATTERN, TROM_ACTIVATE, NULL);
	}
	else
	{
	    /* Copy the search pattern into the string gadget buffer */
	    TR_SetAttribute (RefListPrj, ID_LIST_PATTERN, NULL, (ULONG)args->Reference);
	}

	/* Use the string gadget as pattern buffer from now on */
	args->Reference = (STRPTR) TR_GetAttribute (RefListPrj, ID_LIST_PATTERN, NULL);

	ret->Result = RET_NO_MATCH;
	while (ret->Result == RET_NO_MATCH)
	{
	    TR_Wait (Application, NULL);

	    while (msg = TR_GetMsg (Application))
	    {
		if (msg->trm_Class == TRMS_HELP)
		    GiveHelp (scr);
		else if (msg->trm_Class == TRMS_CLOSEWINDOW)
		    ret->Result = RET_ABORT;
		else
		{
		    switch (msg->trm_ID)
		    {
			/* Listview; if double click we fall through to
			 * the "Fetch reference" gadget, ID_LIST_OK. */
			case ID_LIST_LIST:
			    if (!(msg->trm_Qualifier & IEQUALIFIER_REPEAT))
				break;

			/* "Fetch reference" gadget */
			case ID_LIST_OK:
			{
			    struct ListViewNode *f;

			    TR_LockProject (RefListPrj);

			    /* Find the selected node and fetch it */
			    f = ListOffsetToPtr (&RefsFound, TR_GetAttribute(RefListPrj, ID_LIST_LIST, TRAT_Value));
			    PutRef (f, args->DestFile, ret);

			    TR_UnlockProject (RefListPrj);
			}
			break;

			/* Build a new list based on the new search pattern */
			case ID_LIST_PATTERN:
			    TR_LockProject (RefListPrj);

			    if (BuildRefList (args) == 0)
			    {
				/* Disable fetch and 'list file' gadget if no references are
				 * found.
				 */
				TR_SetAttribute (RefListPrj, ID_LIST_FILE, TRAT_Disabled, TRUE);
				TR_SetAttribute (RefListPrj, ID_LIST_OK, TRAT_Disabled, TRUE);
			    }
			    else
			    {
				/* As we get a new list we can once again do ID_LIST_FILE
				 * and fetch a reference
				 */
				TR_SetAttribute (RefListPrj, ID_LIST_FILE, TRAT_Disabled, FALSE);
				TR_SetAttribute (RefListPrj, ID_LIST_OK, TRAT_Disabled, FALSE);
			    }

			    /* Now actually display the new list */
			    TR_SetAttribute (RefListPrj, ID_LIST_LIST, NULL, (ULONG)&RefsFound);

			    /* And finally re-activate the GUI */
			    TR_UnlockProject (RefListPrj);
			break;

			/* Switch reference list to the file of
			 * the selected reference.
			 */
			case ID_LIST_FILE:
			    SwitchToFileIndex();
			break;

			/* Cancel */
			case ID_LIST_CANCEL:
			    ret->Result = RET_ABORT;
			break;
		    }
		}

		TR_ReplyMsg (msg);
	    }
	}

	TR_CloseProject (RefListPrj);
	RefListPrj = NULL;
    }

    if (swappedscreen)
	ScreenToBack (scr);
    UnlockPubScreen (NULL, scr);
}

/* GiveHelp() - open guide positioned at the help for <id> */
void __regargs GiveHelp (struct Screen *scr)
{
    struct Library *AmigaGuideBase;
    struct NewAmigaGuide nag = { NULL };
    AMIGAGUIDECONTEXT handle;

    nag.nag_Name = "FetchRefs_FR.guide";
    nag.nag_Screen = scr;
    nag.nag_Node = "4.";

    /* Show the guide */
    if (AmigaGuideBase = OpenLibrary ("amigaguide.library", 34))
    {
	if (handle = OpenAmigaGuideA (&nag, NULL))
	    CloseAmigaGuide (handle);
	else
	    PostMessage (GetString (TEXTE_GUIDE));

	CloseLibrary (AmigaGuideBase);
    }
    else
	PostMessage (GetString (TEXTE_AMIGAGUIDE));
}

/* NewMatchList() - free the list (if any) of matches and allocate new one */
APTR NewMatchList (void)
{
    /* Empty the list and create a new, empty one. Empty the listview, too */
    if (RefListPrj)
	TR_SetAttribute (RefListPrj, ID_LIST_LIST, NULL, (ULONG)~0);

    NewList (&RefsFound);

    if (RefListPrj)
    {
	TR_SetAttribute (RefListPrj, ID_LIST_LIST, NULL, (ULONG)&RefsFound);
	TR_SetAttribute (RefListPrj, ID_LIST_LIST, NULL, (ULONG)~0);
    }

    LibDeletePool (ListPool);
    return (ListPool = LibCreatePool (NULL, 8192, 8192));
}

/* PutRef() - writes a reference to file */
/* Write the reference. Handles the writing both to a normal file and to the
 * clipboard (through iffparse.library). */
#define ID_FTXT MAKE_ID('F','T','X','T')
#define ID_CHRS MAKE_ID('C','H','R','S')
static void __regargs PutRef (struct ListViewNode *item, STRPTR destfile, struct FindRefReturnStruct *ret)
{
    BPTR from;
    STRPTR buf = NULL;
    LONG fetchfrom, fetchlength;

    /* Figure out the actual file length if the reference is a complete file.
     * Otherwise just grab the reference size from the list.
     */
    fetchfrom = 0;
    fetchlength = -1;
    if (item->node.ln_Type == 2)
    {
	fetchfrom = item->refsentry->Offset;
	fetchlength = item->refsentry->Length;
    }
    if (fetchlength == -1)
	fetchlength = FileLength (item->fileentry->Name);

    /* Attempt to open file, allocate memory, and read reference */
    if (from = Open (item->fileentry->Name, MODE_OLDFILE))
    {
	if (buf = LibAllocPooled (ListPool, fetchlength))
	{
	    Seek (from, fetchfrom, OFFSET_BEGINNING);
	    Read (from, buf, fetchlength);
	}

	Close (from);
    }
    else
	ret->Number = IoErr();

    /* Now write the reference */
    if (buf)
    {
	/* Write to a normal file if destination is not "CLIP#?" */
	if (strncmp (destfile, "CLIP", 4) != 0)
	{
	    BPTR to;

	    if (to = Open (destfile, MODE_NEWFILE))
	    {
		if (Write (to, buf, fetchlength) == fetchlength)
		    ret->Result = RET_MATCH;
		else
		    ret->Number = IoErr();

		Close (to);
	    } else
		ret->Number = IoErr();

	}
	else  /* Write to a clipboard unit */
	{
	    struct Library *IFFParseBase;
	    if (IFFParseBase = OpenLibrary ("iffparse.library", 37))
	    {
		struct IFFHandle *iff;

		if (iff = AllocIFF())
		{
		    if (iff->iff_Stream = (unsigned long) OpenClipboard (atoi(&(destfile[4]))))
		    {
			InitIFFasClip(iff);
			if (OpenIFF(iff, IFFF_WRITE) == 0)
			{
			    if (PushChunk (iff, ID_FTXT, ID_FORM, IFFSIZE_UNKNOWN) == 0)
			    {
				if (PushChunk (iff, 0, ID_CHRS, fetchlength) == 0)
				{
				    if (WriteChunkBytes (iff, buf, fetchlength) == fetchlength)
				    {
					if (PopChunk (iff) == 0)
					    ret->Result = RET_MATCH;
				    }
				    PopChunk (iff);
				}
			    }
			    CloseIFF (iff);
			}
			CloseClipboard ((struct ClipboardHandle *)iff->iff_Stream);
		    }
		    FreeIFF (iff);
		}
		CloseLibrary (IFFParseBase);
	    }
	}
    }

    /* If we wrote the reference we can return what line number to jump to.
     * For a file reference this is zero.
     */
    if (ret->Result == RET_MATCH)
	ret->Number = (item->node.ln_Type == 1) ? 0 : item->refsentry->Goto;
}

/* BuildRefList() - builds a list of all the matching references */
/* This is a call back function for the sort routine.. */
BOOL __regargs SortListFunct (struct ListViewNode *a, struct ListViewNode *b, ULONG data)
{   BOOL r;
    /* Sort alphabetically */
    if (a->node.ln_Type == b->node.ln_Type)
    {	r = ( stricmp (a->node.ln_Name, b->node.ln_Name) > 0 );
	return r;
    }

    /* Put files at the top */
    r = (a->node.ln_Type == 2);
    return r;
}

static LONG __regargs BuildRefList (struct FindRefOptions *args)
{
    STRPTR pattern;
    LONG numofrefs = 0, len, res;
    struct RefsEntry *currref;
    struct FileEntry *currfile;

    /* First of all: Initialize the list of matching references */
    if (! NewMatchList())
	return ERROR_NO_FREE_STORE;

    /* Tokenize pattern string */
    len = strlen (args->Reference) * 2 + 2;

    /* We just return as fail code is already set above */
    if (! (pattern = LibAllocPooled (ListPool, len)))
	return ERROR_NO_FREE_STORE;

    /* Tokenize the search string */
    if (args->Case)
	res = ParsePattern (args->Reference, pattern, len);
    else
	res = ParsePatternNoCase (args->Reference, pattern, len);

    if (res == -1)
	return ERROR_TOO_MANY_LEVELS;

    /* Pick up all references */
    for (currfile = (struct FileEntry *) FileList.lh_Head;
	    currfile->node.ln_Succ;
	    currfile = (struct FileEntry *) currfile->node.ln_Succ)
    {
	/* Check for a match on the file name */
	if (args->FileRef)
	{
	    UBYTE name[108], *cutptr;
	    /* Isolate the base name (eg. 'INCLUDE:libraries/dos.h' -> 'dos') */
	    strcpy (name, FilePart (currfile->Name));
	    cutptr = name;
	    while (*cutptr)
	    {
		if ((*cutptr >= 'a' && *cutptr <= 'z') ||
		    (*cutptr >= 'A' && *cutptr <= 'Z') ||
		    (*cutptr >= '0' && *cutptr <= '9') ||
		    (*cutptr == '_'))
		    cutptr++;
		else
		    *cutptr = 0;
	    }

	    /* Create an item to add to the RefsFound list */
	    if ((args->Case) ? MatchPattern(pattern, name) : MatchPatternNoCase(pattern, name))
	    {
		struct ListViewNode *r;

		if (r = LibAllocPooled (ListPool, sizeof(struct ListViewNode)))
		{
		    r->fileentry = currfile;
		    r->node.ln_Type = 1;
		    r->node.ln_Name = currfile->Name;

		    AddTail (&RefsFound, &r->node);
		    numofrefs++;
		}
	    }
	}

	/* Check all the RefsEntry's in this FileEntry for a match */
	for (currref = (struct RefsEntry *) currfile->RefsList.lh_Head;
		currref->node.ln_Succ;
		currref = (struct RefsEntry *) currref->node.ln_Succ)
	{
	    if ((args->Case) ? MatchPattern(pattern, currref->Name) : MatchPatternNoCase(pattern, currref->Name))
	    {
		struct ListViewNode *r;
		if (r = LibAllocPooled (ListPool, sizeof(struct ListViewNode) + strlen(currfile->Name) + strlen(currref->Name) + 5))
		{
		    r->fileentry = currfile;
		    r->refsentry = currref;
		    r->node.ln_Type = 2;
		    r->node.ln_Name = r->ListViewText;

		    SPrintf (r->ListViewText, "%s in %s", currref->Name, currfile->Name);
		    /* strcpy (r->ListViewText, currref->Name);
		    strcat (r->ListViewText, " in ");
		    strcat (r->ListViewText, currfile->Name); */

		    AddTail (&RefsFound, &r->node);
		    numofrefs++;
		}
	    }
	}
    }
    SortExecList (&RefsFound, SortListFunct, 0);

    /* Negate the number of matches; positive return code means failure */
    return -numofrefs;
}

static void SwitchToFileIndex (void)
{
    struct FileEntry *file;
    struct RefsEntry *selectedref, *ref;
    struct ListViewNode *selected;
    LONG selectedpos = 0;

    TR_LockProject (RefListPrj);
    TR_SetAttribute (RefListPrj, ID_LIST_FILE, TRAT_Disabled, TRUE);

    selected = ListOffsetToPtr (&RefsFound, TR_GetAttribute(RefListPrj, ID_LIST_LIST, TRAT_Value));
    file = selected->fileentry;
    selectedref = selected->refsentry;

    if (!NewMatchList())
	goto nomem;

    /* Build a new list */
    for (ref = (struct RefsEntry *) file->RefsList.lh_Head;
	    ref->node.ln_Succ;
	    ref = (struct RefsEntry *) ref->node.ln_Succ)
    {
	struct ListViewNode *r;

	if (r = LibAllocPooled (ListPool, sizeof(struct ListViewNode) + strlen(ref->Name) + 4 + strlen(file->Name) + 1))
	{
	    r->fileentry = file;
	    r->refsentry = ref;
	    r->node.ln_Type = 2;
	    r->node.ln_Name = r->ListViewText;
	    SPrintf (r->ListViewText, "%s => %s", ref->Name, file->Name);

	    AddTail (&RefsFound, &r->node);
	}
    }

    SortExecList (&RefsFound, SortListFunct, 0);

    /* Figure out where the reference we started with has ended in new list */
    for (selected = (struct ListViewNode *) RefsFound.lh_Head;
	    selected->node.ln_Succ;
	    selected = (struct ListViewNode *) selected->node.ln_Succ)
    {
	if (selected->refsentry == selectedref)
	    break;
	selectedpos++;
    }
    if (!selected->node.ln_Succ)
	selectedpos = 0;

nomem:
    /* Update GUI */
    TR_SetAttribute (RefListPrj, ID_LIST_LIST, NULL, (ULONG)&RefsFound);
    TR_SetAttribute (RefListPrj, ID_LIST_LIST, TRAT_Value, selectedpos);
    TR_SetAttribute (RefListPrj, ID_LIST_LIST, TRLV_Top, selectedpos);

    TR_UnlockProject (RefListPrj);
}

/* FindScreen() - return best Screen, based prefered <pubname> PubScr name */
static struct Screen * __regargs FindScreen (STRPTR pubname)
{
    struct Screen *s = NULL;

    /* A specific public screen is wanted */
    if (pubname)
	s = LockPubScreen (pubname);

    /* Go for the currently active screen */
    if (!s)
    {
	UBYTE wantname[MAXPUBSCREENNAME+1];
	struct Screen *wantscr;
	struct List *publist;
	struct PubScreenNode *pubnode;

	publist = LockPubScreenList();
	wantscr = IntuitionBase->ActiveScreen;
	pubnode = (struct PubScreenNode *) publist->lh_Head;
	while (pubnode->psn_Node.ln_Succ)
	{
	    if (pubnode->psn_Screen == wantscr)
		strcpy (wantname, pubnode->psn_Node.ln_Name);
	    pubnode = (struct PubScreenNode *) pubnode->psn_Node.ln_Succ;
	}
	UnlockPubScreenList();
	s = LockPubScreen (wantname);
    }

    /* Last option: default public screen */
    if (!s)
	s = LockPubScreen (NULL);

    return s;
}

/* FileLength() - returns file length or -1 for failure */
LONG __regargs FileLength (STRPTR name)
{
    BPTR lock;
    __aligned struct FileInfoBlock fib;

    if (lock = Lock (name, ACCESS_READ))
    {
	Examine (lock, &fib);
	UnLock (lock);
	return fib.fib_Size;
    }
    else
	return -1;
}

void * __regargs ListOffsetToPtr (struct List *l, LONG o)
{
    struct Node *n;

    if (o < 0)
	return NULL;

    n = l->lh_Head;
    while ((o--) && (n->ln_Succ))
	n = n->ln_Succ;

    return (n->ln_Succ) ? n : NULL;
}
